
import math

import numpy as np

from .base import GravitationalWaveTransient
from ...core.utils import (
    logger, speed_of_light, solar_mass, radius_of_earth,
    gravitational_constant, round_up_to_power_of_two
)


class MBGravitationalWaveTransient(GravitationalWaveTransient):
    """A multi-banded likelihood object

    This uses the method described in S. Morisaki, 2021, arXiv: 2104.07813.

    Parameters
    ----------
    interferometers: list, bilby.gw.detector.InterferometerList
        A list of `bilby.detector.Interferometer` instances - contains the detector data and power spectral densities
    waveform_generator: `bilby.waveform_generator.WaveformGenerator`
        An object which computes the frequency-domain strain of the signal, given some set of parameters
    reference_chirp_mass: float
        A reference chirp mass for determining the frequency banding
    highest_mode: int, optional
        The maximum magnetic number of gravitational-wave moments. Default is 2
    linear_interpolation: bool, optional
        If True, the linear-interpolation method is used for the computation of (h, h). If False, the IFFT-FFT method
        is used. Default is True.
    accuracy_factor: float, optional
        A parameter to determine the accuracy of multi-banding. The larger this factor is, the more accurate the
        approximation is. This corresponds to L in the paper. Default is 5.
    time_offset: float, optional
        (end time of data) - (maximum arrival time). If None, it is inferred from the prior of geocent time.
    delta_f_end: float, optional
        The frequency scale with which waveforms at the high-frequency end are smoothed. If None, it is determined from
        the prior of geocent time.
    maximum_banding_frequency: float, optional
        A maximum frequency for multi-banding. If specified, the low-frequency limit of a band does not exceed it.
    minimum_banding_duration: float, optional
        A minimum duration for multi-banding. If specified, the duration of a band is not smaller than it.
    distance_marginalization: bool, optional
        If true, marginalize over distance in the likelihood. This uses a look up table calculated at run time. The
        distance prior is set to be a delta function at the minimum distance allowed in the prior being marginalised
        over.
    phase_marginalization: bool, optional
        If true, marginalize over phase in the likelihood. This is done analytically using a Bessel function. The phase
        prior is set to be a delta function at phase=0.
    priors: dict, bilby.prior.PriorDict
        A dictionary of priors containing at least the geocent_time prior
    distance_marginalization_lookup_table: (dict, str), optional
        If a dict, dictionary containing the lookup_table, distance_array, (distance) prior_array, and
        reference_distance used to construct the table. If a string the name of a file containing these quantities. The
        lookup table is stored after construction in either the provided string or a default location:
        '.distance_marginalization_lookup_dmin{}_dmax{}_n{}.npz'
    reference_frame: (str, bilby.gw.detector.InterferometerList, list), optional
        Definition of the reference frame for the sky location.
        - "sky": sample in RA/dec, this is the default
        - e.g., "H1L1", ["H1", "L1"], InterferometerList(["H1", "L1"]):
          sample in azimuth and zenith, `azimuth` and `zenith` defined in the frame where the z-axis is aligned the the
          vector connecting H1 and L1.
    time_reference: str, optional
        Name of the reference for the sampled time parameter.
        - "geocent"/"geocenter": sample in the time at the Earth's center, this is the default
        - e.g., "H1": sample in the time of arrival at H1

    Returns
    -------
    Likelihood: `bilby.core.likelihood.Likelihood`
        A likelihood object, able to compute the likelihood of the data given some model parameters

    """
    def __init__(
            self, interferometers, waveform_generator, reference_chirp_mass, highest_mode=2, linear_interpolation=True,
            accuracy_factor=5, time_offset=None, delta_f_end=None, maximum_banding_frequency=None,
            minimum_banding_duration=0., distance_marginalization=False, phase_marginalization=False, priors=None,
            distance_marginalization_lookup_table=None, reference_frame="sky", time_reference="geocenter"
    ):
        super(MBGravitationalWaveTransient, self).__init__(
            interferometers=interferometers, waveform_generator=waveform_generator, priors=priors,
            distance_marginalization=distance_marginalization, phase_marginalization=phase_marginalization,
            time_marginalization=False, distance_marginalization_lookup_table=distance_marginalization_lookup_table,
            jitter_time=False, reference_frame=reference_frame, time_reference=time_reference
        )
        self.reference_chirp_mass = reference_chirp_mass
        self.highest_mode = highest_mode
        self.linear_interpolation = linear_interpolation
        self.accuracy_factor = accuracy_factor
        self.time_offset = time_offset
        self.delta_f_end = delta_f_end
        self.minimum_frequency = np.min([i.minimum_frequency for i in self.interferometers])
        self.maximum_frequency = np.max([i.maximum_frequency for i in self.interferometers])
        self.maximum_banding_frequency = maximum_banding_frequency
        self.minimum_banding_duration = minimum_banding_duration
        self.setup_multibanding()

    @property
    def reference_chirp_mass(self):
        return self._reference_chirp_mass

    @property
    def reference_chirp_mass_in_second(self):
        return gravitational_constant * self._reference_chirp_mass * solar_mass / speed_of_light**3.

    @reference_chirp_mass.setter
    def reference_chirp_mass(self, reference_chirp_mass):
        if isinstance(reference_chirp_mass, int) or isinstance(reference_chirp_mass, float):
            self._reference_chirp_mass = reference_chirp_mass
        else:
            raise TypeError("reference_chirp_mass must be a number")

    @property
    def highest_mode(self):
        return self._highest_mode

    @highest_mode.setter
    def highest_mode(self, highest_mode):
        if isinstance(highest_mode, int) or isinstance(highest_mode, float):
            self._highest_mode = highest_mode
        else:
            raise TypeError("highest_mode must be a number")

    @property
    def linear_interpolation(self):
        return self._linear_interpolation

    @linear_interpolation.setter
    def linear_interpolation(self, linear_interpolation):
        if isinstance(linear_interpolation, bool):
            self._linear_interpolation = linear_interpolation
        else:
            raise TypeError("linear_interpolation must be a bool")

    @property
    def accuracy_factor(self):
        return self._accuracy_factor

    @accuracy_factor.setter
    def accuracy_factor(self, accuracy_factor):
        if isinstance(accuracy_factor, int) or isinstance(accuracy_factor, float):
            self._accuracy_factor = accuracy_factor
        else:
            raise TypeError("accuracy_factor must be a number")

    @property
    def time_offset(self):
        return self._time_offset

    @time_offset.setter
    def time_offset(self, time_offset):
        """
        This sets the time offset assumed when frequency bands are constructed. The default value is (the
        maximum offset of geocent time in the prior range) +  (light-traveling time of the Earth). If the
        prior does not contain 'geocent_time', 2.12 seconds is used. It is calculated assuming that the
        maximum offset of geocent time is 2.1 seconds, which is the value for the standard prior used by
        LIGO-Virgo-KAGRA.
        """
        time_parameter = self.time_reference + "_time"
        if time_parameter == "geocent_time":
            safety = radius_of_earth / speed_of_light
        else:
            safety = 2 * radius_of_earth / speed_of_light
        if time_offset is not None:
            if isinstance(time_offset, int) or isinstance(time_offset, float):
                self._time_offset = time_offset
            else:
                raise TypeError("time_offset must be a number")
        elif self.priors is not None and time_parameter in self.priors:
            self._time_offset = (
                self.interferometers.start_time + self.interferometers.duration
                - self.priors[time_parameter].minimum + safety
            )
        else:
            self._time_offset = 2.12
            logger.warning("time offset can not be inferred. Use the standard time offset of {} seconds.".format(
                self._time_offset))

    @property
    def delta_f_end(self):
        return self._delta_f_end

    @delta_f_end.setter
    def delta_f_end(self, delta_f_end):
        """
        This sets the frequency scale of tapering the high-frequency end of waveform, to avoid the issues of
        abrupt termination of waveform described in Sec. 2. F of arXiv: 2104.07813. This needs to be much
        larger than the inverse of the minimum time offset, and the default value is 100 times of that. If
        the prior does not contain 'geocent_time' and the minimum time offset can not be computed, 53Hz is
        used. It is computed assuming that the minimum offset of geocent time is 1.9 seconds, which is the
        value for the standard prior used by LIGO-Virgo-KAGRA.
        """
        time_parameter = self.time_reference + "_time"
        if time_parameter == "geocent_time":
            safety = radius_of_earth / speed_of_light
        else:
            safety = 2 * radius_of_earth / speed_of_light
        if delta_f_end is not None:
            if isinstance(delta_f_end, int) or isinstance(delta_f_end, float):
                self._delta_f_end = delta_f_end
            else:
                raise TypeError("delta_f_end must be a number")
        elif self.priors is not None and time_parameter in self.priors:
            self._delta_f_end = 100 / (
                self.interferometers.start_time + self.interferometers.duration
                - self.priors[time_parameter].maximum - safety
            )
        else:
            self._delta_f_end = 53.
            logger.warning("delta_f_end can not be inferred. Use the standard delta_f_end of {} Hz.".format(
                self._delta_f_end))

    @property
    def maximum_banding_frequency(self):
        return self._maximum_banding_frequency

    @maximum_banding_frequency.setter
    def maximum_banding_frequency(self, maximum_banding_frequency):
        r"""
        This sets the upper limit on a starting frequency of a band. The default value is the frequency at
        which f - 1 / \sqrt(- d\tau / df) starts to decrease, because the bisection search of the starting
        frequency does not work from that frequency. The stationary phase approximation is not valid at such
        a high frequency, which can break down the approximation. It is calculated from the 0PN formula of
        time-to-merger \tau(f). The user-specified frequency is used if it is lower than that frequency.
        """
        fmax_tmp = (
            (15 / 968)**(3 / 5) * (self.highest_mode / (2 * np.pi))**(8 / 5)
            / self.reference_chirp_mass_in_second
        )
        if maximum_banding_frequency is not None:
            if isinstance(maximum_banding_frequency, int) or isinstance(maximum_banding_frequency, float):
                if maximum_banding_frequency < fmax_tmp:
                    fmax_tmp = maximum_banding_frequency
                else:
                    logger.warning("The input maximum_banding_frequency is too large."
                                   "It is set to be {} Hz.".format(fmax_tmp))
            else:
                raise TypeError("maximum_banding_frequency must be a number")
        self._maximum_banding_frequency = fmax_tmp

    @property
    def minimum_banding_duration(self):
        return self._minimum_banding_duration

    @minimum_banding_duration.setter
    def minimum_banding_duration(self, minimum_banding_duration):
        if isinstance(minimum_banding_duration, int) or isinstance(minimum_banding_duration, float):
            self._minimum_banding_duration = minimum_banding_duration
        else:
            raise TypeError("minimum_banding_duration must be a number")

    def setup_multibanding(self):
        """Set up frequency bands and coefficients needed for likelihood evaluations"""
        self._setup_frequency_bands()
        self._setup_integers()
        self._setup_waveform_frequency_points()
        self._setup_linear_coefficients()
        if self.linear_interpolation:
            self._setup_quadratic_coefficients_linear_interp()
        else:
            self._setup_quadratic_coefficients_ifft_fft()

    def _tau(self, f):
        """Compute time-to-merger from the input frequency. This uses the 0PN formula.

        Parameters
        ----------
        f: float
            input frequency

        Returns
        -------
        tau: float
            time-to-merger

        """
        f_22 = 2 * f / self.highest_mode
        return (
            5 / 256 * self.reference_chirp_mass_in_second
            * (np.pi * self.reference_chirp_mass_in_second * f_22) ** (-8 / 3)
        )

    def _dtaudf(self, f):
        """Compute the derivative of time-to-merger with respect to a starting frequency. This uses the 0PN formula.

        Parameters
        ----------
        f: float
            input frequency

        Returns
        -------
        dtaudf: float
            derivative of time-to-merger

        """
        f_22 = 2 * f / self.highest_mode
        return (
            -5 / 96 * self.reference_chirp_mass_in_second
            * (np.pi * self.reference_chirp_mass_in_second * f_22) ** (-8. / 3.) / f
        )

    def _find_starting_frequency(self, duration, fnow):
        """Find the starting frequency of the next band satisfying (10) and
        (51) of arXiv: 2104.07813.

        Parameters
        ----------
        duration: float
            duration of the next band
        fnow: float
            starting frequency of the current band

        Returns
        -------
        fnext: float or None
            starting frequency of the next band. None if a frequency satisfying the conditions does not exist.
        dfnext: float or None
            frequency scale with which waveforms are smoothed. None if a frequency satisfying the conditions does not
            exist.

        """
        def _is_above_fnext(f):
            """This function returns True if f > fnext"""
            cond1 = (
                duration - self.time_offset - self._tau(f)
                - self.accuracy_factor * np.sqrt(-self._dtaudf(f))
            ) > 0
            cond2 = f - 1. / np.sqrt(-self._dtaudf(f)) - fnow > 0
            return cond1 and cond2
        # Bisection search for fnext
        fmin, fmax = fnow, self.maximum_banding_frequency
        if not _is_above_fnext(fmax):
            return None, None
        while fmax - fmin > 1e-2 / duration:
            f = (fmin + fmax) / 2.
            if _is_above_fnext(f):
                fmax = f
            else:
                fmin = f
        return f, 1. / np.sqrt(-self._dtaudf(f))

    def _setup_frequency_bands(self):
        r"""Set up frequency bands. The durations of bands geometrically decrease T, T/2. T/4, ..., where T is the
        original duration. This sets the following instance variables.

        durations: durations of bands (T^(b) in the paper)
        fb_dfb: the list of tuples, which contain starting frequencies (f^(b) in the paper) and frequency scales for
        smoothing waveforms (\Delta f^(b) in the paper) of bands

        """
        self.durations = [self.interferometers.duration]
        self.fb_dfb = [(self.minimum_frequency, 0.)]
        dnext = self.interferometers.duration / 2
        while dnext > max(self.time_offset, self.minimum_banding_duration):
            fnow, _ = self.fb_dfb[-1]
            fnext, dfnext = self._find_starting_frequency(dnext, fnow)
            if fnext is not None and fnext < min(self.maximum_frequency, self.maximum_banding_frequency):
                self.durations.append(dnext)
                self.fb_dfb.append((fnext, dfnext))
                dnext /= 2
            else:
                break
        self.fb_dfb.append((self.maximum_frequency + self.delta_f_end, self.delta_f_end))
        logger.info("The total frequency range is divided into {} bands with frequency intervals of {}.".format(
            len(self.durations), ", ".join(["1/{} Hz".format(d) for d in self.durations])))

    def _setup_integers(self):
        """Set up integers needed for likelihood evaluations. This sets the following instance variables.

        Nbs: the numbers of samples of downsampled data (N^(b) in the paper)
        Mbs: the numbers of samples of shortened data (M^(b) in the paper)
        Ks_Ke: start and end frequency indices of bands (K^(b)_s and K^(b)_e in the paper)

        """
        self.Nbs = []
        self.Mbs = []
        self.Ks_Ke = []
        for b in range(len(self.durations)):
            dnow = self.durations[b]
            fnow, dfnow = self.fb_dfb[b]
            fnext, _ = self.fb_dfb[b + 1]
            Nb = max(round_up_to_power_of_two(2. * (fnext * self.interferometers.duration + 1.)), 2**b)
            self.Nbs.append(Nb)
            self.Mbs.append(Nb // 2**b)
            self.Ks_Ke.append((math.ceil((fnow - dfnow) * dnow), math.floor(fnext * dnow)))

    def _setup_waveform_frequency_points(self):
        """Set up frequency points where waveforms are evaluated. Frequency points are reordered because some waveform
        models raise an error if the input frequencies are not increasing. This adds frequency_points into the
        waveform_arguments of waveform_generator. This sets the following instance variables.

        banded_frequency_points: ndarray of total banded frequency points
        start_end_idxs: list of tuples containing start and end indices of each band
        unique_to_original_frequencies: indices converting unique frequency
        points into the original duplicated banded frequencies

        """
        self.banded_frequency_points = np.array([])
        self.start_end_idxs = []
        start_idx = 0
        for i in range(len(self.fb_dfb) - 1):
            d = self.durations[i]
            Ks, Ke = self.Ks_Ke[i]
            self.banded_frequency_points = np.append(self.banded_frequency_points, np.arange(Ks, Ke + 1) / d)
            end_idx = start_idx + Ke - Ks
            self.start_end_idxs.append((start_idx, end_idx))
            start_idx = end_idx + 1
        unique_frequencies, idxs = np.unique(self.banded_frequency_points, return_inverse=True)
        self.waveform_generator.waveform_arguments['frequencies'] = unique_frequencies
        self.unique_to_original_frequencies = idxs
        logger.info("The number of frequency points where waveforms are evaluated is {}.".format(
            len(unique_frequencies)))
        logger.info("The speed-up gain of multi-banding is {}.".format(
            (self.maximum_frequency - self.minimum_frequency) * self.interferometers.duration /
            len(unique_frequencies)))

    def _window(self, f, b):
        """Compute window function in the b-th band

        Parameters
        ----------
        f: float or ndarray
            frequency at which the window function is computed
        b: int

        Returns
        -------
        window: float
            window function at f
        """
        fnow, dfnow = self.fb_dfb[b]
        fnext, dfnext = self.fb_dfb[b + 1]

        @np.vectorize
        def _vectorized_window(f):
            if fnow - dfnow < f < fnow:
                return (1. + np.cos(np.pi * (f - fnow) / dfnow)) / 2.
            elif fnow <= f <= fnext - dfnext:
                return 1.
            elif fnext - dfnext < f < fnext:
                return (1. - np.cos(np.pi * (f - fnext) / dfnext)) / 2.
            else:
                return 0.

        return _vectorized_window(f)

    def _setup_linear_coefficients(self):
        """Set up coefficients by which waveforms are multiplied to compute (d, h)"""
        self.linear_coeffs = dict((ifo.name, np.array([])) for ifo in self.interferometers)
        N = self.Nbs[-1]
        for ifo in self.interferometers:
            logger.info("Pre-computing linear coefficients for {}".format(ifo.name))
            fddata = np.zeros(N // 2 + 1, dtype=complex)
            fddata[:len(ifo.frequency_domain_strain)][ifo.frequency_mask] += \
                ifo.frequency_domain_strain[ifo.frequency_mask] / ifo.power_spectral_density_array[ifo.frequency_mask]
            for b in range(len(self.fb_dfb) - 1):
                start_idx, end_idx = self.start_end_idxs[b]
                windows = self._window(self.banded_frequency_points[start_idx:end_idx + 1], b)
                fddata_in_ith_band = np.copy(fddata[:int(self.Nbs[b] / 2 + 1)])
                fddata_in_ith_band[-1] = 0.  # zeroing data at the Nyquist frequency
                tddata = np.fft.irfft(fddata_in_ith_band)[-self.Mbs[b]:]
                Ks, Ke = self.Ks_Ke[b]
                fddata_in_ith_band = np.fft.rfft(tddata)[Ks:Ke + 1]
                self.linear_coeffs[ifo.name] = np.append(
                    self.linear_coeffs[ifo.name], (4. / self.durations[b]) * windows * np.conj(fddata_in_ith_band))

    def _setup_quadratic_coefficients_linear_interp(self):
        """Set up coefficients by which the squares of waveforms are multiplied to compute (h, h) for the
        linear-interpolation algorithm"""
        logger.info("Linear-interpolation algorithm is used for (h, h).")
        self.quadratic_coeffs = dict((ifo.name, np.array([])) for ifo in self.interferometers)
        N = self.Nbs[-1]
        for ifo in self.interferometers:
            logger.info("Pre-computing quadratic coefficients for {}".format(ifo.name))
            full_frequencies = np.arange(N // 2 + 1) / ifo.duration
            full_inv_psds = np.zeros(N // 2 + 1)
            full_inv_psds[:len(ifo.power_spectral_density_array)][ifo.frequency_mask] = \
                1. / ifo.power_spectral_density_array[ifo.frequency_mask]
            for i in range(len(self.fb_dfb) - 1):
                start_idx, end_idx = self.start_end_idxs[i]
                banded_frequencies = self.banded_frequency_points[start_idx:end_idx + 1]
                coeffs = np.zeros(len(banded_frequencies))
                for k in range(len(coeffs) - 1):
                    if k == 0:
                        start_idx_in_sum = 0
                    else:
                        start_idx_in_sum = math.ceil(ifo.duration * banded_frequencies[k])
                    if k == len(coeffs) - 2:
                        end_idx_in_sum = len(full_frequencies) - 1
                    else:
                        end_idx_in_sum = math.ceil(ifo.duration * banded_frequencies[k + 1]) - 1
                    window_over_psd = (
                        full_inv_psds[start_idx_in_sum:end_idx_in_sum + 1]
                        * self._window(full_frequencies[start_idx_in_sum:end_idx_in_sum + 1], i)
                    )
                    frequencies_in_sum = full_frequencies[start_idx_in_sum:end_idx_in_sum + 1]
                    coeffs[k] += 4 * self.durations[i] / ifo.duration * np.sum(
                        (banded_frequencies[k + 1] - frequencies_in_sum) * window_over_psd
                    )
                    coeffs[k + 1] += 4 * self.durations[i] / ifo.duration * np.sum(
                        (frequencies_in_sum - banded_frequencies[k]) * window_over_psd
                    )
                self.quadratic_coeffs[ifo.name] = np.append(self.quadratic_coeffs[ifo.name], coeffs)

    def _setup_quadratic_coefficients_ifft_fft(self):
        """Set up coefficients needed for the IFFT-FFT algorithm to compute (h, h)"""
        logger.info("IFFT-FFT algorithm is used for (h, h).")
        N = self.Nbs[-1]
        # variables defined below correspond to \hat{N}^(b), \hat{T}^(b), \tilde{I}^(b)_{c, k}, h^(b)_{c, m} and
        # \sqrt{w^(b)(f^(b)_k)} \tilde{h}(f^(b)_k) in the paper
        Nhatbs = [min(2 * Mb, Nb) for Mb, Nb in zip(self.Mbs, self.Nbs)]
        self.Tbhats = [self.interferometers.duration * Nbhat / Nb for Nb, Nbhat in zip(self.Nbs, Nhatbs)]
        self.Ibcs = dict((ifo.name, []) for ifo in self.interferometers)
        self.hbcs = dict((ifo.name, []) for ifo in self.interferometers)
        self.wths = dict((ifo.name, []) for ifo in self.interferometers)
        for ifo in self.interferometers:
            logger.info("Pre-computing quadratic coefficients for {}".format(ifo.name))
            full_inv_psds = np.zeros(N // 2 + 1)
            full_inv_psds[:len(ifo.power_spectral_density_array)][ifo.frequency_mask] = (
                1 / ifo.power_spectral_density_array[ifo.frequency_mask]
            )
            for b in range(len(self.fb_dfb) - 1):
                Imb = np.fft.irfft(full_inv_psds[:self.Nbs[b] // 2 + 1])
                half_length = Nhatbs[b] // 2
                Imbc = np.append(Imb[:half_length + 1], Imb[-(Nhatbs[b] - half_length - 1):])
                self.Ibcs[ifo.name].append(np.fft.rfft(Imbc))
                # Allocate arrays for IFFT-FFT operations
                self.hbcs[ifo.name].append(np.zeros(Nhatbs[b]))
                self.wths[ifo.name].append(np.zeros(self.Mbs[b] // 2 + 1, dtype=complex))
        # precompute windows and their squares
        self.windows = np.array([])
        self.square_root_windows = np.array([])
        for b in range(len(self.fb_dfb) - 1):
            start, end = self.start_end_idxs[b]
            ws = self._window(self.banded_frequency_points[start:end + 1], b)
            self.windows = np.append(self.windows, ws)
            self.square_root_windows = np.append(self.square_root_windows, np.sqrt(ws))

    def calculate_snrs(self, waveform_polarizations, interferometer):
        """
        Compute the snrs for multi-banding

        Parameters
        ----------
        waveform_polarizations: waveform
        interferometer: bilby.gw.detector.Interferometer

        Returns
        -------
        snrs: named tuple of snrs

        """
        strain = np.zeros(len(self.banded_frequency_points), dtype=complex)
        for mode in waveform_polarizations:
            response = interferometer.antenna_response(
                self.parameters['ra'], self.parameters['dec'],
                self.parameters['geocent_time'], self.parameters['psi'],
                mode
            )
            strain += waveform_polarizations[mode][self.unique_to_original_frequencies] * response

        dt = interferometer.time_delay_from_geocenter(
            self.parameters['ra'], self.parameters['dec'],
            self.parameters['geocent_time'])
        dt_geocent = self.parameters['geocent_time'] - interferometer.strain_data.start_time
        ifo_time = dt_geocent + dt

        calib_factor = interferometer.calibration_model.get_calibration_factor(
            self.banded_frequency_points, prefix='recalib_{}_'.format(interferometer.name), **self.parameters)

        strain *= np.exp(-1j * 2. * np.pi * self.banded_frequency_points * ifo_time)
        strain *= calib_factor

        d_inner_h = np.dot(strain, self.linear_coeffs[interferometer.name])

        if self.linear_interpolation:
            optimal_snr_squared = np.vdot(
                np.real(strain * np.conjugate(strain)),
                self.quadratic_coeffs[interferometer.name]
            )
        else:
            optimal_snr_squared = 0.
            for b in range(len(self.fb_dfb) - 1):
                Ks, Ke = self.Ks_Ke[b]
                start_idx, end_idx = self.start_end_idxs[b]
                Mb = self.Mbs[b]
                if b == 0:
                    optimal_snr_squared += (4. / self.interferometers.duration) * np.vdot(
                        np.real(strain[start_idx:end_idx + 1] * np.conjugate(strain[start_idx:end_idx + 1])),
                        interferometer.frequency_mask[Ks:Ke + 1] * self.windows[start_idx:end_idx + 1]
                        / interferometer.power_spectral_density_array[Ks:Ke + 1])
                else:
                    self.wths[interferometer.name][b][Ks:Ke + 1] = (
                        self.square_root_windows[start_idx:end_idx + 1] * strain[start_idx:end_idx + 1]
                    )
                    self.hbcs[interferometer.name][b][-Mb:] = np.fft.irfft(self.wths[interferometer.name][b])
                    thbc = np.fft.rfft(self.hbcs[interferometer.name][b])
                    optimal_snr_squared += (4. / self.Tbhats[b]) * np.vdot(
                        np.real(thbc * np.conjugate(thbc)), self.Ibcs[interferometer.name][b])

        complex_matched_filter_snr = d_inner_h / (optimal_snr_squared**0.5)

        return self._CalculatedSNRs(
            d_inner_h=d_inner_h, optimal_snr_squared=optimal_snr_squared,
            complex_matched_filter_snr=complex_matched_filter_snr,
            d_inner_h_squared_tc_array=None,
            d_inner_h_array=None,
            optimal_snr_squared_array=None)

    def _rescale_signal(self, signal, new_distance):
        for mode in signal:
            signal[mode] *= self._ref_dist / new_distance



import unittest

import numpy as np
import bilby


class TestCategoricalPrior(unittest.TestCase):
    def test_single_sample(self):
        categorical_prior = bilby.core.prior.Categorical(3)
        in_prior = True
        for _ in range(1000):
            s = categorical_prior.sample()
            if s not in [0, 1, 2]:
                in_prior = False
        self.assertTrue(in_prior)

    def test_array_sample(self):
        ncat = 4
        categorical_prior = bilby.core.prior.Categorical(ncat)
        N = 100000
        s = categorical_prior.sample(N)
        zeros = np.sum(s == 0)
        ones = np.sum(s == 1)
        twos = np.sum(s == 2)
        threes = np.sum(s == 3)
        self.assertEqual(zeros + ones + twos + threes, N)
        self.assertAlmostEqual(zeros / N, 1 / ncat, places=int(np.log10(np.sqrt(N))))
        self.assertAlmostEqual(ones / N, 1 / ncat, places=int(np.log10(np.sqrt(N))))
        self.assertAlmostEqual(twos / N, 1 / ncat, places=int(np.log10(np.sqrt(N))))
        self.assertAlmostEqual(threes / N, 1 / ncat, places=int(np.log10(np.sqrt(N))))

    def test_single_probability(self):
        N = 3
        categorical_prior = bilby.core.prior.Categorical(N)
        self.assertEqual(categorical_prior.prob(0), 1 / N)
        self.assertEqual(categorical_prior.prob(1), 1 / N)
        self.assertEqual(categorical_prior.prob(2), 1 / N)
        self.assertEqual(categorical_prior.prob(0.5), 0)

    def test_array_probability(self):
        N = 3
        categorical_prior = bilby.core.prior.Categorical(N)
        self.assertTrue(
            np.all(
                categorical_prior.prob([0, 1, 1, 2, 3])
                == np.array([1 / N, 1 / N, 1 / N, 1 / N, 0])
            )
        )

    def test_single_lnprobability(self):
        N = 3
        categorical_prior = bilby.core.prior.Categorical(N)
        self.assertEqual(categorical_prior.ln_prob(0), -np.log(N))
        self.assertEqual(categorical_prior.ln_prob(1), -np.log(N))
        self.assertEqual(categorical_prior.ln_prob(2), -np.log(N))
        self.assertEqual(categorical_prior.ln_prob(0.5), -np.inf)

    def test_array_lnprobability(self):
        N = 3
        categorical_prior = bilby.core.prior.Categorical(N)
        self.assertTrue(
            np.all(
                categorical_prior.ln_prob([0, 1, 1, 2, 3])
                == np.array([-np.log(N), -np.log(N), -np.log(N), -np.log(N), -np.inf])
            )
        )


if __name__ == "__main__":
    unittest.main()



from collections import namedtuple

LOGLKEY = "logl"
LOGLLATEXKEY = r"$\log\mathcal{L}$"
LOGPKEY = "logp"
LOGPLATEXKEY = r"$\log\pi$"

ConvergenceInputs = namedtuple(
    "ConvergenceInputs",
    [
        "autocorr_c",
        "burn_in_nact",
        "thin_by_nact",
        "fixed_discard",
        "target_nsamples",
        "stop_after_convergence",
        "L1steps",
        "L2steps",
        "min_tau",
        "fixed_tau",
        "tau_window",
    ],
)

ParallelTemperingInputs = namedtuple(
    "ParallelTemperingInputs",
    [
        "ntemps",
        "nensemble",
        "Tmax",
        "Tmax_from_SNR",
        "initial_betas",
        "adapt",
        "adapt_t0",
        "adapt_nu",
        "pt_ensemble",
    ],
)



import numpy as np
from pandas import DataFrame

from .base_sampler import NestedSampler, signal_wrapper


class Nestle(NestedSampler):
    """bilby wrapper `nestle.Sampler` (http://kylebarbary.com/nestle/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `nestle.sample`, see documentation for
    that function for further help. Under Other Parameters, we list commonly
    used kwargs and the bilby defaults

    Parameters
    ==========
    npoints: int
        The number of live points, note this can also equivalently be given as
        one of [nlive, nlives, n_live_points]
    method: {'classic', 'single', 'multi'} ('multi')
        Method used to select new points
    verbose: Bool
        If true, print information information about the convergence during
        sampling

    """

    default_kwargs = dict(
        verbose=True,
        method="multi",
        npoints=500,
        update_interval=None,
        npdim=None,
        maxiter=None,
        maxcall=None,
        dlogz=None,
        decline_factor=None,
        rstate=None,
        callback=None,
        steps=20,
        enlarge=1.2,
    )

    def _translate_kwargs(self, kwargs):
        if "npoints" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["npoints"] = kwargs.pop(equiv)
        if "steps" not in kwargs:
            for equiv in self.walks_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["steps"] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        if self.kwargs["verbose"]:
            import nestle

            self.kwargs["callback"] = nestle.print_progress
            self.kwargs.pop("verbose")
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    @signal_wrapper
    def run_sampler(self):
        """Runs Nestle sampler with given kwargs and returns the result

        Returns
        =======
        bilby.core.result.Result: Packaged information about the result

        """
        import nestle

        out = nestle.sample(
            loglikelihood=self.log_likelihood,
            prior_transform=self.prior_transform,
            ndim=self.ndim,
            **self.kwargs
        )
        print("")

        self.result.sampler_output = out
        self.result.samples = nestle.resample_equal(out.samples, out.weights)
        self.result.nested_samples = DataFrame(
            out.samples, columns=self.search_parameter_keys
        )
        self.result.nested_samples["weights"] = out.weights
        self.result.nested_samples["log_likelihood"] = out.logl
        self.result.log_likelihood_evaluations = self.reorder_loglikelihoods(
            unsorted_loglikelihoods=out.logl,
            unsorted_samples=out.samples,
            sorted_samples=self.result.samples,
        )
        self.result.log_evidence = out.logz
        self.result.log_evidence_err = out.logzerr
        self.result.information_gain = out.h
        self.calc_likelihood_count()
        return self.result

    def _run_test(self):
        """
        Runs to test whether the sampler is properly running with the given
        kwargs without actually running to the end

        Returns
        =======
        bilby.core.result.Result: Dummy container for sampling results.

        """
        import nestle

        kwargs = self.kwargs.copy()
        kwargs["maxiter"] = 2
        nestle.sample(
            loglikelihood=self.log_likelihood,
            prior_transform=self.prior_transform,
            ndim=self.ndim,
            **kwargs
        )
        self.result.samples = np.random.uniform(0, 1, (100, self.ndim))
        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan
        self.calc_likelihood_count()
        return self.result

    def write_current_state(self):
        """
        Nestle doesn't support checkpointing so no current state will be
        written on interrupt.
        """
        pass



import functools
import os
from distutils.spawn import find_executable

from .log import logger


def latex_plot_format(func):
    """
    Wrap the plotting function to set rcParams dependent on environment variables

    The rcparams can be set directly from the env. variable `BILBY_STYLE` to
    point to a matplotlib style file. Or, if `BILBY_STYLE=default` (any case) a
    default setup is used, this is enabled by default. To not use any rcParams,
    set `BILBY_STYLE=none`. Occasionally, issues arrise with the latex
    `mathdefault` command. A fix is to define this command in the rcParams. An
    env. variable `BILBY_MATHDEFAULT` can be used to turn this fix on/off.
    Setting `BILBY_MATHDEFAULT=1` will enable the fix, all other choices
    (including undefined) will disable it. Additionally, the BILBY_STYLE and
    BILBY_MATHDEFAULT arguments can be passed into any
    latex_plot_format-wrapped plotting function and will be set directly.

    """
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        import matplotlib.pyplot as plt
        from matplotlib import rcParams

        if "BILBY_STYLE" in kwargs:
            bilby_style = kwargs.pop("BILBY_STYLE")
        else:
            bilby_style = os.environ.get("BILBY_STYLE", "default")

        if "BILBY_MATHDEFAULT" in kwargs:
            bilby_mathdefault = kwargs.pop("BILBY_MATHDEFAULT")
        else:
            bilby_mathdefault = int(os.environ.get("BILBY_MATHDEFAULT", "0"))

        if bilby_mathdefault == 1:
            logger.debug("Setting mathdefault in the rcParams")
            rcParams['text.latex.preamble'] = r'\providecommand{\mathdefault}[1][]{}'

        logger.debug("Using BILBY_STYLE={}".format(bilby_style))
        if bilby_style.lower() == "none":
            return func(*args, **kwargs)
        elif os.path.isfile(bilby_style):
            plt.style.use(bilby_style)
            return func(*args, **kwargs)
        elif bilby_style in plt.style.available:
            plt.style.use(bilby_style)
            return func(*args, **kwargs)
        elif bilby_style.lower() == "default":
            _old_tex = rcParams["text.usetex"]
            _old_serif = rcParams["font.serif"]
            _old_family = rcParams["font.family"]
            if find_executable("latex"):
                rcParams["text.usetex"] = True
            else:
                rcParams["text.usetex"] = False
            rcParams["font.serif"] = "Computer Modern Roman"
            rcParams["font.family"] = "serif"
            rcParams["text.usetex"] = _old_tex
            rcParams["font.serif"] = _old_serif
            rcParams["font.family"] = _old_family
            return func(*args, **kwargs)
        else:
            logger.debug(
                "Environment variable BILBY_STYLE={} not used"
                .format(bilby_style)
            )
            return func(*args, **kwargs)
    return wrapper_decorator



import numpy as np
from scipy.special import logsumexp


class SamplesSummary(object):
    """ Object to store a set of samples and calculate summary statistics

    Parameters
    ==========
    samples: array_like
        Array of samples
    average: str {'median', 'mean'}
        Use either a median average or mean average when calculating relative
        uncertainties
    level: float
        The default confidence interval level, defaults t0 0.9

    """
    def __init__(self, samples, average='median', confidence_level=.9):
        self.samples = samples
        self.average = average
        self.confidence_level = confidence_level

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, samples):
        self._samples = samples

    @property
    def confidence_level(self):
        return self._confidence_level

    @confidence_level.setter
    def confidence_level(self, confidence_level):
        if 0 < confidence_level and confidence_level < 1:
            self._confidence_level = confidence_level
        else:
            raise ValueError("Confidence level must be between 0 and 1")

    @property
    def average(self):
        if self._average == 'mean':
            return self.mean
        elif self._average == 'median':
            return self.median

    @average.setter
    def average(self, average):
        allowed_averages = ['mean', 'median']
        if average in allowed_averages:
            self._average = average
        else:
            raise ValueError("Average {} not in allowed averages".format(average))

    @property
    def median(self):
        return np.median(self.samples, axis=0)

    @property
    def mean(self):
        return np.mean(self.samples, axis=0)

    @property
    def _lower_level(self):
        """ The credible interval lower quantile value """
        return (1 - self.confidence_level) / 2.

    @property
    def _upper_level(self):
        """ The credible interval upper quantile value """
        return (1 + self.confidence_level) / 2.

    @property
    def lower_absolute_credible_interval(self):
        """ Absolute lower value of the credible interval """
        return np.quantile(self.samples, self._lower_level, axis=0)

    @property
    def upper_absolute_credible_interval(self):
        """ Absolute upper value of the credible interval """
        return np.quantile(self.samples, self._upper_level, axis=0)

    @property
    def lower_relative_credible_interval(self):
        """ Relative (to average) lower value of the credible interval """
        return self.lower_absolute_credible_interval - self.average

    @property
    def upper_relative_credible_interval(self):
        """ Relative (to average) upper value of the credible interval """
        return self.upper_absolute_credible_interval - self.average


def kish_log_effective_sample_size(ln_weights):
    """ Calculate the Kish effective sample size from the natural-log weights

    See https://en.wikipedia.org/wiki/Effective_sample_size for details

    Parameters
    ==========
    ln_weights: array
        An array of the ln-weights

    Returns
    =======
    ln_n_eff:
        The natural-log of the effective sample size

    """
    return 2 * logsumexp(ln_weights) - logsumexp(2 * ln_weights)


def reflect(u):
    """
    Iteratively reflect a number until it is contained in [0, 1].

    This is for priors with a reflective boundary condition, all numbers in the
    set `u = 2n +/- x` should be mapped to x.

    For the `+` case we just take `u % 1`.
    For the `-` case we take `1 - (u % 1)`.

    E.g., -0.9, 1.1, and 2.9 should all map to 0.9.

    Parameters
    ==========
    u: array-like
        The array of points to map to the unit cube

    Returns
    =======
    u: array-like
       The input array, modified in place.
    """
    idxs_even = np.mod(u, 2) < 1
    u[idxs_even] = np.mod(u[idxs_even], 1)
    u[~idxs_even] = 1 - np.mod(u[~idxs_even], 1)
    return u



import matplotlib

matplotlib.use("Agg")  # noqa

import glob
import importlib.util
import unittest
import os
import parameterized
import pytest
import shutil
import logging

import bilby.core.utils

bilby.core.utils.command_line_args.clean = True
core_examples = glob.glob("examples/core_examples/*.py")
core_examples += glob.glob("examples/core_examples/*/*.py")
core_args = [(fname.split("/")[-1][:-3], fname) for fname in core_examples]

gw_examples = [
    "examples/gw_examples/injection_examples/fast_tutorial.py",
    "examples/gw_examples/data_examples/GW150914.py",
]
gw_args = [(fname.split("/")[-1][:-3], fname) for fname in gw_examples]


def _execute_file(name, fname):
    dname, fname = os.path.split(fname)
    old_directory = os.getcwd()
    os.chdir(dname)
    spec = importlib.util.spec_from_file_location(name, fname)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    os.chdir(old_directory)


class ExampleTest(unittest.TestCase):
    outdir = "outdir"
    dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.path.abspath(os.path.join(dir_path, os.path.pardir))

    def setUp(self):
        self.init_dir = os.getcwd()

    def tearDown(self):
        if os.path.isdir(self.outdir):
            try:
                shutil.rmtree(self.outdir)
            except OSError:
                logging.warning("{} not removed after tests".format(self.outdir))
        os.chdir(self.init_dir)

    @classmethod
    def setUpClass(cls):
        if os.path.isdir(cls.outdir):
            try:
                shutil.rmtree(cls.outdir)
            except OSError:
                logging.warning("{} not removed prior to tests".format(cls.outdir))

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir(cls.outdir):
            try:
                shutil.rmtree(cls.outdir)
            except OSError:
                logging.warning("{} not removed prior to tests".format(cls.outdir))

    @parameterized.parameterized.expand(core_args)
    def test_core_examples(self, name, fname):
        """ Loop over examples to check they run """
        bilby.core.utils.command_line_args.bilby_test_mode = False
        ignore = ["15d_gaussian"]
        if any([item in fname for item in ignore]):
            pytest.skip()
        _execute_file(name, fname)

    @parameterized.parameterized.expand(gw_args)
    def test_gw_examples(self, name, fname):
        """ Loop over examples to check they run """
        bilby.core.utils.command_line_args.bilby_test_mode = True
        _execute_file(name, fname)


if __name__ == "__main__":
    unittest.main()



import datetime
import importlib
import os
import time

import numpy as np

from ..utils import logger
from .base_sampler import NestedSampler, _TemporaryFileSamplerMixin, signal_wrapper


class Pymultinest(_TemporaryFileSamplerMixin, NestedSampler):
    """
    bilby wrapper of pymultinest
    (https://github.com/JohannesBuchner/PyMultiNest)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `pymultinest.run`, see documentation
    for that class for further help. Under Other Parameters, we list commonly
    used kwargs and the bilby defaults.

    Parameters
    ==========
    npoints: int
        The number of live points, note this can also equivalently be given as
        one of [nlive, nlives, n_live_points]
    importance_nested_sampling: bool, (False)
        If true, use importance nested sampling
    sampling_efficiency: float or {'parameter', 'model'}, ('parameter')
        Defines the sampling efficiency
    verbose: Bool
        If true, print information information about the convergence during
    resume: bool
        If true, resume run from checkpoint (if available)

    """

    default_kwargs = dict(
        importance_nested_sampling=False,
        resume=True,
        verbose=True,
        sampling_efficiency="parameter",
        n_live_points=500,
        n_params=None,
        n_clustering_params=None,
        wrapped_params=None,
        multimodal=True,
        const_efficiency_mode=False,
        evidence_tolerance=0.5,
        n_iter_before_update=100,
        null_log_evidence=-1e90,
        max_modes=100,
        mode_tolerance=-1e90,
        outputfiles_basename=None,
        seed=-1,
        context=0,
        write_output=True,
        log_zero=-1e100,
        max_iter=0,
        init_MPI=False,
        dump_callback=None,
    )
    short_name = "pm"
    hard_exit = True

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        exit_code=77,
        skip_import_verification=False,
        temporary_directory=True,
        **kwargs
    ):
        try:
            from mpi4py import MPI

            using_mpi = MPI.COMM_WORLD.Get_size() > 1
        except ImportError:
            using_mpi = False
        super(Pymultinest, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            exit_code=exit_code,
            temporary_directory=temporary_directory,
            **kwargs
        )
        self._apply_multinest_boundaries()
        self.exit_code = exit_code
        if using_mpi and temporary_directory:
            logger.info(
                "Temporary directory incompatible with MPI, "
                "will run in original directory"
            )
        self.use_temporary_directory = temporary_directory and not using_mpi

    def _translate_kwargs(self, kwargs):
        if "n_live_points" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["n_live_points"] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        """Check the kwargs"""

        self.outputfiles_basename = self.kwargs.pop("outputfiles_basename", None)

        # for PyMultiNest >=2.9 the n_params kwarg cannot be None
        if self.kwargs["n_params"] is None:
            self.kwargs["n_params"] = self.ndim
        if self.kwargs["dump_callback"] is None:
            self.kwargs["dump_callback"] = self._dump_callback
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    def _dump_callback(self, *args, **kwargs):
        if self.use_temporary_directory:
            self._copy_temporary_directory_contents_to_proper_path()
        self._calculate_and_save_sampling_time()

    def _apply_multinest_boundaries(self):
        if self.kwargs["wrapped_params"] is None:
            self.kwargs["wrapped_params"] = list()
            for param in self.search_parameter_keys:
                if self.priors[param].boundary == "periodic":
                    self.kwargs["wrapped_params"].append(1)
                else:
                    self.kwargs["wrapped_params"].append(0)

    @signal_wrapper
    def run_sampler(self):
        import pymultinest

        self._verify_kwargs_against_default_kwargs()

        self._setup_run_directory()
        self._check_and_load_sampling_time_file()

        # Overwrite pymultinest's signal handling function
        pm_run = importlib.import_module("pymultinest.run")
        pm_run.interrupt_handler = self.write_current_state_and_exit

        self.start_time = time.time()
        out = pymultinest.solve(
            LogLikelihood=self.log_likelihood,
            Prior=self.prior_transform,
            n_dims=self.ndim,
            **self.kwargs
        )
        self._calculate_and_save_sampling_time()

        self._clean_up_run_directory()

        post_equal_weights = os.path.join(
            self.outputfiles_basename, "post_equal_weights.dat"
        )
        post_equal_weights_data = np.loadtxt(post_equal_weights)
        self.result.log_likelihood_evaluations = post_equal_weights_data[:, -1]
        self.result.sampler_output = out
        self.result.samples = post_equal_weights_data[:, :-1]
        self.result.log_evidence = out["logZ"]
        self.result.log_evidence_err = out["logZerr"]
        self.calc_likelihood_count()
        self.result.outputfiles_basename = self.outputfiles_basename
        self.result.sampling_time = datetime.timedelta(seconds=self.total_sampling_time)
        self.result.nested_samples = self._nested_samples
        return self.result

    @property
    def _nested_samples(self):
        """
        Extract nested samples from the pymultinest files.
        This requires combining the "dead" points from `ev.dat` and the "live"
        points from `phys_live.points`.
        The prior volume associated with the current live points is the simple
        estimate of `remaining_prior_volume / N`.
        """
        import pandas as pd

        dir_ = self.kwargs["outputfiles_basename"]
        dead_points = np.genfromtxt(dir_ + "/ev.dat")
        live_points = np.genfromtxt(dir_ + "/phys_live.points")

        nlive = self.kwargs["n_live_points"]
        final_log_prior_volume = -len(dead_points) / nlive - np.log(nlive)
        live_points = np.insert(live_points, -1, final_log_prior_volume, axis=-1)

        nested_samples = pd.DataFrame(
            np.vstack([dead_points, live_points]).copy(),
            columns=self.search_parameter_keys
            + ["log_likelihood", "log_prior_volume", "mode"],
        )
        return nested_samples



import os

import numpy as np
from pandas import DataFrame

from ..utils import check_directory_exists_and_if_not_mkdir, load_json, logger
from .base_sampler import NestedSampler


class Nessai(NestedSampler):
    """bilby wrapper of nessai (https://github.com/mj-will/nessai)

    All positional and keyword arguments passed to `run_sampler` are propagated
    to `nessai.flowsampler.FlowSampler`

    See the documentation for an explanation of the different kwargs.

    Documentation: https://nessai.readthedocs.io/
    """

    _default_kwargs = None
    seed_equiv_kwargs = ["sampling_seed"]

    @property
    def default_kwargs(self):
        """Default kwargs for nessai.

        Retrieves default values from nessai directly and then includes any
        bilby specific defaults. This avoids the need to update bilby when the
        defaults change or new kwargs are added to nessai.
        """
        if not self._default_kwargs:
            from inspect import signature

            from nessai.flowsampler import FlowSampler
            from nessai.nestedsampler import NestedSampler
            from nessai.proposal import AugmentedFlowProposal, FlowProposal

            kwargs = {}
            classes = [
                AugmentedFlowProposal,
                FlowProposal,
                NestedSampler,
                FlowSampler,
            ]
            for c in classes:
                kwargs.update(
                    {
                        k: v.default
                        for k, v in signature(c).parameters.items()
                        if v.default is not v.empty
                    }
                )
            # Defaults for bilby that will override nessai defaults
            bilby_defaults = dict(output=None, exit_code=self.exit_code)
            kwargs.update(bilby_defaults)
            self._default_kwargs = kwargs
        return self._default_kwargs

    def log_prior(self, theta):
        """

        Parameters
        ----------
        theta: list
            List of sampled values on a unit interval

        Returns
        -------
        float: Joint ln prior probability of theta

        """
        return self.priors.ln_prob(theta, axis=0)

    def run_sampler(self):
        from nessai.flowsampler import FlowSampler
        from nessai.livepoint import dict_to_live_points, live_points_to_array
        from nessai.model import Model as BaseModel
        from nessai.posterior import compute_weights
        from nessai.utils import setup_logger

        class Model(BaseModel):
            """A wrapper class to pass our log_likelihood and priors into nessai

            Parameters
            ----------
            names : list of str
                List of parameters to sample
            priors : :obj:`bilby.core.prior.PriorDict`
                Priors to use for sampling. Needed for the bounds and the
                `sample` method.
            """

            def __init__(self, names, priors):
                self.names = names
                self.priors = priors
                self._update_bounds()

            @staticmethod
            def log_likelihood(x, **kwargs):
                """Compute the log likelihood"""
                theta = [x[n].item() for n in self.search_parameter_keys]
                return self.log_likelihood(theta)

            @staticmethod
            def log_prior(x, **kwargs):
                """Compute the log prior"""
                theta = {n: x[n] for n in self._search_parameter_keys}
                return self.log_prior(theta)

            def _update_bounds(self):
                self.bounds = {
                    key: [self.priors[key].minimum, self.priors[key].maximum]
                    for key in self.names
                }

            def new_point(self, N=1):
                """Draw a point from the prior"""
                prior_samples = self.priors.sample(size=N)
                samples = {n: prior_samples[n] for n in self.names}
                return dict_to_live_points(samples)

            def new_point_log_prob(self, x):
                """Proposal probability for new the point"""
                return self.log_prior(x)

        # Setup the logger for nessai using the same settings as the bilby logger
        setup_logger(
            self.outdir, label=self.label, log_level=logger.getEffectiveLevel()
        )
        model = Model(self.search_parameter_keys, self.priors)
        try:
            out = FlowSampler(model, **self.kwargs)
            out.run(save=True, plot=self.plot)
        except TypeError as e:
            raise TypeError(f"Unable to initialise nessai sampler with error: {e}")
        except (SystemExit, KeyboardInterrupt) as e:
            import sys

            logger.info(
                f"Caught {type(e).__name__} with args {e.args}, "
                f"exiting with signal {self.exit_code}"
            )
            sys.exit(self.exit_code)

        # Manually set likelihood evaluations because parallelisation breaks the counter
        self.result.num_likelihood_evaluations = out.ns.likelihood_evaluations[-1]

        self.result.samples = live_points_to_array(
            out.posterior_samples, self.search_parameter_keys
        )
        self.result.log_likelihood_evaluations = out.posterior_samples["logL"]
        self.result.nested_samples = DataFrame(out.nested_samples)
        self.result.nested_samples.rename(
            columns=dict(logL="log_likelihood", logP="log_prior"), inplace=True
        )
        _, log_weights = compute_weights(
            np.array(self.result.nested_samples.log_likelihood),
            np.array(out.ns.state.nlive),
        )
        self.result.nested_samples["weights"] = np.exp(log_weights)
        self.result.log_evidence = out.ns.log_evidence
        self.result.log_evidence_err = np.sqrt(out.ns.information / out.ns.nlive)

        return self.result

    def _translate_kwargs(self, kwargs):
        if "nlive" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nlive"] = kwargs.pop(equiv)
        if "n_pool" not in kwargs:
            for equiv in self.npool_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["n_pool"] = kwargs.pop(equiv)
            if "n_pool" not in kwargs:
                kwargs["n_pool"] = self._npool
        if "seed" not in kwargs:
            for equiv in self.seed_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["seed"] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        """
        Set the directory where the output will be written
        and check resume and checkpoint status.
        """
        if "config_file" in self.kwargs:
            d = load_json(self.kwargs["config_file"], None)
            self.kwargs.update(d)
            self.kwargs.pop("config_file")

        if not self.kwargs["plot"]:
            self.kwargs["plot"] = self.plot

        if self.kwargs["n_pool"] == 1 and self.kwargs["max_threads"] == 1:
            logger.warning("Setting pool to None (n_pool=1 & max_threads=1)")
            self.kwargs["n_pool"] = None

        if not self.kwargs["output"]:
            self.kwargs["output"] = os.path.join(
                self.outdir, f"{self.label}_nessai", ""
            )

        check_directory_exists_and_if_not_mkdir(self.kwargs["output"])
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    def _setup_pool(self):
        pass



import datetime

import numpy as np

from ..utils import logger
from .base_sampler import Sampler, signal_wrapper
from .dynesty import Dynesty, _log_likelihood_wrapper, _prior_transform_wrapper


class DynamicDynesty(Dynesty):
    """
    bilby wrapper of `dynesty.DynamicNestedSampler`
    (https://dynesty.readthedocs.io/en/latest/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `dynesty.DynamicNestedSampler`, see
    documentation for that class for further help. Under Other Parameter below,
    we list commonly all kwargs and the bilby defaults.

    Parameters
    ==========
    likelihood: likelihood.Likelihood
        A  object with a log_l method
    priors: bilby.core.prior.PriorDict, dict
        Priors to be used in the search.
        This has attributes for each parameter to be sampled.
    outdir: str, optional
        Name of the output directory
    label: str, optional
        Naming scheme of the output files
    use_ratio: bool, optional
        Switch to set whether or not you want to use the log-likelihood ratio
        or just the log-likelihood
    plot: bool, optional
        Switch to set whether or not you want to create traceplots
    skip_import_verification: bool
        Skips the check if the sampler is installed if true. This is
        only advisable for testing environments

    Other Parameters
    ------==========
    bound: {'none', 'single', 'multi', 'balls', 'cubes'}, ('multi')
        Method used to select new points
    sample: {'unif', 'rwalk', 'slice', 'rslice', 'hslice'}, ('rwalk')
        Method used to sample uniformly within the likelihood constraints,
        conditioned on the provided bounds
    walks: int
        Number of walks taken if using `sample='rwalk'`, defaults to `ndim * 5`
    verbose: Bool
        If true, print information information about the convergence during
    check_point: bool,
        If true, use check pointing.
    check_point_delta_t: float (600)
        The approximate checkpoint period (in seconds). Should the run be
        interrupted, it can be resumed from the last checkpoint. Set to
        `None` to turn-off check pointing
    n_check_point: int, optional (None)
        The number of steps to take before check pointing (override
        check_point_delta_t).
    resume: bool
        If true, resume run from checkpoint (if available)
    """

    default_kwargs = dict(
        bound="multi",
        sample="rwalk",
        verbose=True,
        check_point_delta_t=600,
        first_update=None,
        npdim=None,
        rstate=None,
        queue_size=None,
        pool=None,
        use_pool=None,
        logl_args=None,
        logl_kwargs=None,
        ptform_args=None,
        ptform_kwargs=None,
        enlarge=None,
        bootstrap=None,
        vol_dec=0.5,
        vol_check=2.0,
        facc=0.5,
        slices=5,
        walks=None,
        update_interval=0.6,
        nlive_init=500,
        maxiter_init=None,
        maxcall_init=None,
        dlogz_init=0.01,
        logl_max_init=np.inf,
        nlive_batch=500,
        wt_function=None,
        wt_kwargs=None,
        maxiter_batch=None,
        maxcall_batch=None,
        maxiter=None,
        maxcall=None,
        maxbatch=None,
        stop_function=None,
        stop_kwargs=None,
        use_stop=True,
        save_bounds=True,
        print_progress=True,
        print_func=None,
        live_points=None,
    )

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        skip_import_verification=False,
        check_point=True,
        n_check_point=None,
        check_point_delta_t=600,
        resume=True,
        **kwargs,
    ):
        super(DynamicDynesty, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            **kwargs,
        )
        self.n_check_point = n_check_point
        self.check_point = check_point
        self.resume = resume
        if self.n_check_point is None:
            # If the log_likelihood_eval_time is not calculable then
            # check_point is set to False.
            if np.isnan(self._log_likelihood_eval_time):
                self.check_point = False
            n_check_point_raw = check_point_delta_t / self._log_likelihood_eval_time
            n_check_point_rnd = int(float(f"{n_check_point_raw:1.0g}"))
            self.n_check_point = n_check_point_rnd

        self.resume_file = f"{self.outdir}/{self.label}_resume.pickle"

    @property
    def external_sampler_name(self):
        return "dynesty"

    @property
    def sampler_function_kwargs(self):
        keys = [
            "nlive_init",
            "maxiter_init",
            "maxcall_init",
            "dlogz_init",
            "logl_max_init",
            "nlive_batch",
            "wt_function",
            "wt_kwargs",
            "maxiter_batch",
            "maxcall_batch",
            "maxiter",
            "maxcall",
            "maxbatch",
            "stop_function",
            "stop_kwargs",
            "use_stop",
            "save_bounds",
            "print_progress",
            "print_func",
            "live_points",
        ]
        return {key: self.kwargs[key] for key in keys}

    @signal_wrapper
    def run_sampler(self):
        import dynesty

        self._setup_pool()
        self.sampler = dynesty.DynamicNestedSampler(
            loglikelihood=_log_likelihood_wrapper,
            prior_transform=_prior_transform_wrapper,
            ndim=self.ndim,
            **self.sampler_init_kwargs,
        )

        if self.check_point:
            out = self._run_external_sampler_with_checkpointing()
        else:
            out = self._run_external_sampler_without_checkpointing()
        self._close_pool()

        # Flushes the output to force a line break
        if self.kwargs["verbose"]:
            print("")

        # self.result.sampler_output = out
        self._generate_result(out)
        if self.plot:
            self.generate_trace_plots(out)

        return self.result

    def _run_external_sampler_with_checkpointing(self):
        logger.debug("Running sampler with checkpointing")
        if self.resume:
            resume = self.read_saved_state(continuing=True)
            if resume:
                logger.info("Resuming from previous run.")

        old_ncall = self.sampler.ncall
        sampler_kwargs = self.sampler_function_kwargs.copy()
        sampler_kwargs["maxcall"] = self.n_check_point
        self.start_time = datetime.datetime.now()
        while True:
            sampler_kwargs["maxcall"] += self.n_check_point
            self.sampler.run_nested(**sampler_kwargs)
            if self.sampler.ncall == old_ncall:
                break
            old_ncall = self.sampler.ncall

            self.write_current_state()

        self._remove_checkpoint()
        return self.sampler.results

    def write_current_state_and_exit(self, signum=None, frame=None):
        Sampler.write_current_state_and_exit(self=self, signum=signum, frame=frame)

    def _verify_kwargs_against_default_kwargs(self):
        Sampler._verify_kwargs_against_default_kwargs(self)



import numpy as np

from ...core import utils
from ...core.series import CoupledTimeAndFrequencySeries
from ...core.utils import logger, PropertyAccessor
from .. import utils as gwutils


class InterferometerStrainData(object):
    """ Strain data for an interferometer """

    duration = PropertyAccessor('_times_and_frequencies', 'duration')
    sampling_frequency = PropertyAccessor('_times_and_frequencies', 'sampling_frequency')
    start_time = PropertyAccessor('_times_and_frequencies', 'start_time')
    frequency_array = PropertyAccessor('_times_and_frequencies', 'frequency_array')
    time_array = PropertyAccessor('_times_and_frequencies', 'time_array')

    def __init__(self, minimum_frequency=0, maximum_frequency=np.inf,
                 roll_off=0.2, notch_list=None):
        """ Initiate an InterferometerStrainData object

        The initialised object contains no data, this should be added using one
        of the `set_from..` methods.

        Parameters
        ==========
        minimum_frequency: float
            Minimum frequency to analyse for detector. Default is 0.
        maximum_frequency: float
            Maximum frequency to analyse for detector. Default is infinity.
        roll_off: float
            The roll-off (in seconds) used in the Tukey window, default=0.2s.
            This corresponds to alpha * duration / 2 for scipy tukey window.
        notch_list: bilby.gw.detector.strain_data.NotchList
            A list of notches

        """

        self.minimum_frequency = minimum_frequency
        self.maximum_frequency = maximum_frequency
        self.notch_list = notch_list
        self.roll_off = roll_off
        self.window_factor = 1

        self._times_and_frequencies = CoupledTimeAndFrequencySeries()

        self._frequency_mask_updated = False
        self._frequency_mask = None
        self._frequency_domain_strain = None
        self._time_domain_strain = None
        self._channel = None

    def __eq__(self, other):
        if self.minimum_frequency == other.minimum_frequency \
                and self.maximum_frequency == other.maximum_frequency \
                and self.roll_off == other.roll_off \
                and self.window_factor == other.window_factor \
                and self.sampling_frequency == other.sampling_frequency \
                and self.duration == other.duration \
                and self.start_time == other.start_time \
                and np.array_equal(self.time_array, other.time_array) \
                and np.array_equal(self.frequency_array, other.frequency_array) \
                and np.array_equal(self.frequency_domain_strain, other.frequency_domain_strain) \
                and np.array_equal(self.time_domain_strain, other.time_domain_strain):
            return True
        return False

    def time_within_data(self, time):
        """ Check if time is within the data span

        Parameters
        ==========
        time: float
            The time to check

        Returns
        =======
        bool:
            A boolean stating whether the time is inside or outside the span

        """
        if time < self.start_time:
            logger.debug("Time is before the start_time")
            return False
        elif time > self.start_time + self.duration:
            logger.debug("Time is after the start_time + duration")
            return False
        else:
            return True

    @property
    def minimum_frequency(self):
        return self._minimum_frequency

    @minimum_frequency.setter
    def minimum_frequency(self, minimum_frequency):
        self._minimum_frequency = minimum_frequency
        self._frequency_mask_updated = False

    @property
    def maximum_frequency(self):
        """ Force the maximum frequency be less than the Nyquist frequency """
        if self.sampling_frequency is not None:
            if 2 * self._maximum_frequency > self.sampling_frequency:
                self._maximum_frequency = self.sampling_frequency / 2.
        return self._maximum_frequency

    @maximum_frequency.setter
    def maximum_frequency(self, maximum_frequency):
        self._maximum_frequency = maximum_frequency
        self._frequency_mask_updated = False

    @property
    def notch_list(self):
        return self._notch_list

    @notch_list.setter
    def notch_list(self, notch_list):
        """ Set the notch_list

        Parameters
        ==========
        notch_list: list, bilby.gw.detector.strain_data.NotchList
            A list of length-2 tuples of the (max, min) frequency for the
            notches or a pre-made bilby NotchList.

        """
        if notch_list is None:
            self._notch_list = NotchList(None)
        elif isinstance(notch_list, list):
            self._notch_list = NotchList(notch_list)
        elif isinstance(notch_list, NotchList):
            self._notch_list = notch_list
        else:
            raise ValueError("notch_list {} not understood".format(notch_list))
        self._frequency_mask_updated = False

    @property
    def frequency_mask(self):
        """ Masking array for limiting the frequency band.

        Returns
        =======
        mask: np.ndarray
            An array of boolean values
        """
        if not self._frequency_mask_updated:
            frequency_array = self._times_and_frequencies.frequency_array
            mask = ((frequency_array >= self.minimum_frequency) &
                    (frequency_array <= self.maximum_frequency))
            for notch in self.notch_list:
                mask[notch.get_idxs(frequency_array)] = False
            self._frequency_mask = mask
            self._frequency_mask_updated = True
        return self._frequency_mask

    @frequency_mask.setter
    def frequency_mask(self, mask):
        self._frequency_mask = mask
        self._frequency_mask_updated = True

    @property
    def alpha(self):
        return 2 * self.roll_off / self.duration

    def time_domain_window(self, roll_off=None, alpha=None):
        """
        Window function to apply to time domain data before FFTing.

        This defines self.window_factor as the power loss due to the windowing.
        See https://dcc.ligo.org/DocDB/0027/T040089/000/T040089-00.pdf

        Parameters
        ==========
        roll_off: float
            Rise time of window in seconds
        alpha: float
            Parameter to pass to tukey window, how much of segment falls
            into windowed part

        Returns
        =======
        window: array
            Window function over time array
        """
        from scipy.signal.windows import tukey
        if roll_off is not None:
            self.roll_off = roll_off
        elif alpha is not None:
            self.roll_off = alpha * self.duration / 2
        window = tukey(len(self._time_domain_strain), alpha=self.alpha)
        self.window_factor = np.mean(window ** 2)
        return window

    @property
    def time_domain_strain(self):
        """ The time domain strain, in units of strain """
        if self._time_domain_strain is not None:
            return self._time_domain_strain
        elif self._frequency_domain_strain is not None:
            self._time_domain_strain = utils.infft(
                self.frequency_domain_strain, self.sampling_frequency)
            return self._time_domain_strain

        else:
            raise ValueError("time domain strain data not yet set")

    @property
    def frequency_domain_strain(self):
        """ Returns the frequency domain strain

        This is the frequency domain strain normalised to units of
        strain / Hz, obtained by a one-sided Fourier transform of the
        time domain data, divided by the sampling frequency.
        """
        if self._frequency_domain_strain is not None:
            return self._frequency_domain_strain * self.frequency_mask
        elif self._time_domain_strain is not None:
            logger.info("Generating frequency domain strain from given time "
                        "domain strain.")
            logger.info("Applying a tukey window with alpha={}, roll off={}".format(
                self.alpha, self.roll_off))
            # self.low_pass_filter()
            window = self.time_domain_window()
            self._frequency_domain_strain, self.frequency_array = utils.nfft(
                self._time_domain_strain * window, self.sampling_frequency)
            return self._frequency_domain_strain * self.frequency_mask
        else:
            raise ValueError("frequency domain strain data not yet set")

    @frequency_domain_strain.setter
    def frequency_domain_strain(self, frequency_domain_strain):
        if not len(self.frequency_array) == len(frequency_domain_strain):
            raise ValueError("The frequency_array and the set strain have different lengths")
        self._frequency_domain_strain = frequency_domain_strain
        self._time_domain_strain = None

    def to_gwpy_timeseries(self):
        """
        Output the time series strain data as a :class:`gwpy.timeseries.TimeSeries`.
        """
        try:
            from gwpy.timeseries import TimeSeries
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Cannot output strain data as gwpy TimeSeries")

        return TimeSeries(
            self.time_domain_strain, sample_rate=self.sampling_frequency,
            t0=self.start_time, channel=self.channel
        )

    def to_pycbc_timeseries(self):
        """
        Output the time series strain data as a :class:`pycbc.types.timeseries.TimeSeries`.
        """

        try:
            from pycbc.types.timeseries import TimeSeries
            from lal import LIGOTimeGPS
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Cannot output strain data as PyCBC TimeSeries")

        return TimeSeries(
            self.time_domain_strain, delta_t=(1. / self.sampling_frequency),
            epoch=LIGOTimeGPS(self.start_time)
        )

    def to_lal_timeseries(self):
        """
        Output the time series strain data as a LAL TimeSeries object.
        """
        try:
            from lal import CreateREAL8TimeSeries, LIGOTimeGPS, SecondUnit
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Cannot output strain data as PyCBC TimeSeries")

        lal_data = CreateREAL8TimeSeries(
            "", LIGOTimeGPS(self.start_time), 0, 1 / self.sampling_frequency,
            SecondUnit, len(self.time_domain_strain)
        )
        lal_data.data.data[:] = self.time_domain_strain

        return lal_data

    def to_gwpy_frequencyseries(self):
        """
        Output the frequency series strain data as a :class:`gwpy.frequencyseries.FrequencySeries`.
        """
        try:
            from gwpy.frequencyseries import FrequencySeries
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Cannot output strain data as gwpy FrequencySeries")

        return FrequencySeries(
            self.frequency_domain_strain,
            frequencies=self.frequency_array,
            epoch=self.start_time,
            channel=self.channel
        )

    def to_pycbc_frequencyseries(self):
        """
        Output the frequency series strain data as a :class:`pycbc.types.frequencyseries.FrequencySeries`.
        """

        try:
            from pycbc.types.frequencyseries import FrequencySeries
            from lal import LIGOTimeGPS
        except ImportError:
            raise ImportError("Cannot output strain data as PyCBC FrequencySeries")

        return FrequencySeries(
            self.frequency_domain_strain,
            delta_f=1 / self.duration,
            epoch=LIGOTimeGPS(self.start_time)
        )

    def to_lal_frequencyseries(self):
        """
        Output the frequency series strain data as a LAL FrequencySeries object.
        """
        try:
            from lal import CreateCOMPLEX16FrequencySeries, LIGOTimeGPS, SecondUnit
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Cannot output strain data as PyCBC TimeSeries")

        lal_data = CreateCOMPLEX16FrequencySeries(
            "",
            LIGOTimeGPS(self.start_time),
            self.frequency_array[0],
            1 / self.duration,
            SecondUnit,
            len(self.frequency_domain_strain)
        )
        lal_data.data.data[:] = self.frequency_domain_strain

        return lal_data

    def low_pass_filter(self, filter_freq=None):
        """ Low pass filter the data """
        from gwpy.signal.filter_design import lowpass
        from gwpy.timeseries import TimeSeries

        if filter_freq is None:
            logger.debug(
                "Setting low pass filter_freq using given maximum frequency")
            filter_freq = self.maximum_frequency

        if 2 * filter_freq >= self.sampling_frequency:
            logger.info(
                "Low pass filter frequency of {}Hz requested, this is equal"
                " or greater than the Nyquist frequency so no filter applied"
                .format(filter_freq))
            return

        logger.debug("Applying low pass filter with filter frequency {}".format(filter_freq))
        bp = lowpass(filter_freq, self.sampling_frequency)
        strain = TimeSeries(self.time_domain_strain, sample_rate=self.sampling_frequency)
        strain = strain.filter(bp, filtfilt=True)
        self._time_domain_strain = strain.value

    def create_power_spectral_density(
            self, fft_length, overlap=0, name='unknown', outdir=None,
            analysis_segment_start_time=None):
        """ Use the time domain strain to generate a power spectral density

        This create a Tukey-windowed power spectral density and writes it to a
        PSD file.

        Parameters
        ==========
        fft_length: float
            Duration of the analysis segment.
        overlap: float
            Number of seconds of overlap between FFTs.
        name: str
            The name of the detector, used in storing the PSD. Defaults to
            "unknown".
        outdir: str
            The output directory to write the PSD file too. If not given,
            the PSD will not be written to file.
        analysis_segment_start_time: float
            The start time of the analysis segment, if given, this data will
            be removed before creating the PSD.

        Returns
        =======
        frequency_array, psd : array_like
            The frequencies and power spectral density array

        """
        from gwpy.timeseries import TimeSeries

        data = self.time_domain_strain

        if analysis_segment_start_time is not None:
            analysis_segment_end_time = analysis_segment_start_time + fft_length
            inside = (analysis_segment_start_time > self.time_array[0] +
                      analysis_segment_end_time < self.time_array[-1])
            if inside:
                logger.info("Removing analysis segment data from the PSD data")
                idxs = (
                    (self.time_array < analysis_segment_start_time) +
                    (self.time_array > analysis_segment_end_time))
                data = data[idxs]

        # WARNING this line can cause issues if the data is non-contiguous
        strain = TimeSeries(data=data, sample_rate=self.sampling_frequency)
        psd_alpha = 2 * self.roll_off / fft_length
        logger.info(
            "Tukey window PSD data with alpha={}, roll off={}".format(
                psd_alpha, self.roll_off))
        psd = strain.psd(
            fftlength=fft_length, overlap=overlap, window=('tukey', psd_alpha))

        if outdir:
            psd_file = '{}/{}_PSD_{}_{}.txt'.format(outdir, name, self.start_time, self.duration)
            with open('{}'.format(psd_file), 'w+') as opened_file:
                for f, p in zip(psd.frequencies.value, psd.value):
                    opened_file.write('{} {}\n'.format(f, p))

        return psd.frequencies.value, psd.value

    def _infer_time_domain_dependence(
            self, start_time, sampling_frequency, duration, time_array):
        """ Helper function to figure out if the time_array, or
            sampling_frequency and duration where given
        """
        self._infer_dependence(domain='time', array=time_array, duration=duration,
                               sampling_frequency=sampling_frequency, start_time=start_time)

    def _infer_frequency_domain_dependence(
            self, start_time, sampling_frequency, duration, frequency_array):
        """ Helper function to figure out if the frequency_array, or
            sampling_frequency and duration where given
        """

        self._infer_dependence(domain='frequency', array=frequency_array,
                               duration=duration, sampling_frequency=sampling_frequency, start_time=start_time)

    def _infer_dependence(self, domain, array, duration, sampling_frequency, start_time):
        if (sampling_frequency is not None) and (duration is not None):
            if array is not None:
                raise ValueError(
                    "You have given the sampling_frequency, duration, and "
                    "an array")
            pass
        elif array is not None:
            if domain == 'time':
                self.time_array = array
            elif domain == 'frequency':
                self.frequency_array = array
                self.start_time = start_time
            return
        elif sampling_frequency is None or duration is None:
            raise ValueError(
                "You must provide both sampling_frequency and duration")
        else:
            raise ValueError(
                "Insufficient information given to set arrays")
        self._times_and_frequencies = CoupledTimeAndFrequencySeries(duration=duration,
                                                                    sampling_frequency=sampling_frequency,
                                                                    start_time=start_time)

    def set_from_time_domain_strain(
            self, time_domain_strain, sampling_frequency=None, duration=None,
            start_time=0, time_array=None):
        """ Set the strain data from a time domain strain array

        This sets the time_domain_strain attribute, the frequency_domain_strain
        is automatically calculated after a low-pass filter and Tukey window
        is applied.

        Parameters
        ==========
        time_domain_strain: array_like
            An array of the time domain strain.
        sampling_frequency: float
            The sampling frequency (in Hz).
        duration: float
            The data duration (in s).
        start_time: float
            The GPS start-time of the data.
        time_array: array_like
            The array of times, if sampling_frequency and duration not
            given.

        """
        self._infer_time_domain_dependence(start_time=start_time,
                                           sampling_frequency=sampling_frequency,
                                           duration=duration,
                                           time_array=time_array)

        logger.debug('Setting data using provided time_domain_strain')
        if np.shape(time_domain_strain) == np.shape(self.time_array):
            self._time_domain_strain = time_domain_strain
            self._frequency_domain_strain = None
        else:
            raise ValueError("Data times do not match time array")

    def set_from_gwpy_timeseries(self, time_series):
        """ Set the strain data from a gwpy TimeSeries

        This sets the time_domain_strain attribute, the frequency_domain_strain
        is automatically calculated after a low-pass filter and Tukey window
        is applied.

        Parameters
        ==========
        time_series: gwpy.timeseries.timeseries.TimeSeries
            The data to use

        """
        from gwpy.timeseries import TimeSeries
        logger.debug('Setting data using provided gwpy TimeSeries object')
        if not isinstance(time_series, TimeSeries):
            raise ValueError("Input time_series is not a gwpy TimeSeries")
        self._times_and_frequencies = \
            CoupledTimeAndFrequencySeries(duration=time_series.duration.value,
                                          sampling_frequency=time_series.sample_rate.value,
                                          start_time=time_series.epoch.value)
        self._time_domain_strain = time_series.value
        self._frequency_domain_strain = None
        self._channel = time_series.channel

    @property
    def channel(self):
        return self._channel

    def set_from_open_data(
            self, name, start_time, duration=4, outdir='outdir', cache=True,
            **kwargs):
        """ Set the strain data from open LOSC data

        This sets the time_domain_strain attribute, the frequency_domain_strain
        is automatically calculated after a low-pass filter and Tukey window
        is applied.

        Parameters
        ==========
        name: str
            Detector name, e.g., 'H1'.
        start_time: float
            Start GPS time of segment.
        duration: float, optional
            The total time (in seconds) to analyse. Defaults to 4s.
        outdir: str
            Directory where the psd files are saved
        cache: bool, optional
            Whether or not to store/use the acquired data.
        **kwargs:
            All keyword arguments are passed to
            `gwpy.timeseries.TimeSeries.fetch_open_data()`.

        """

        timeseries = gwutils.get_open_strain_data(
            name, start_time, start_time + duration, outdir=outdir, cache=cache,
            **kwargs)

        self.set_from_gwpy_timeseries(timeseries)

    def set_from_csv(self, filename):
        """ Set the strain data from a csv file

        Parameters
        ==========
        filename: str
            The path to the file to read in

        """
        from gwpy.timeseries import TimeSeries
        timeseries = TimeSeries.read(filename, format='csv')
        self.set_from_gwpy_timeseries(timeseries)

    def set_from_frequency_domain_strain(
            self, frequency_domain_strain, sampling_frequency=None,
            duration=None, start_time=0, frequency_array=None):
        """ Set the `frequency_domain_strain` from a numpy array

        Parameters
        ==========
        frequency_domain_strain: array_like
            The data to set.
        sampling_frequency: float
            The sampling frequency (in Hz).
        duration: float
            The data duration (in s).
        start_time: float
            The GPS start-time of the data.
        frequency_array: array_like
            The array of frequencies, if sampling_frequency and duration not
            given.

        """

        self._infer_frequency_domain_dependence(start_time=start_time,
                                                sampling_frequency=sampling_frequency,
                                                duration=duration,
                                                frequency_array=frequency_array)

        logger.debug('Setting data using provided frequency_domain_strain')
        if np.shape(frequency_domain_strain) == np.shape(self.frequency_array):
            self._frequency_domain_strain = frequency_domain_strain
            self.window_factor = 1
        else:
            raise ValueError("Data frequencies do not match frequency_array")

    def set_from_power_spectral_density(
            self, power_spectral_density, sampling_frequency, duration,
            start_time=0):
        """ Set the `frequency_domain_strain` by generating a noise realisation

        Parameters
        ==========
        power_spectral_density: bilby.gw.detector.PowerSpectralDensity
            A PowerSpectralDensity object used to generate the data
        sampling_frequency: float
            The sampling frequency (in Hz)
        duration: float
            The data duration (in s)
        start_time: float
            The GPS start-time of the data

        """

        self._times_and_frequencies = CoupledTimeAndFrequencySeries(duration=duration,
                                                                    sampling_frequency=sampling_frequency,
                                                                    start_time=start_time)
        logger.debug(
            'Setting data using noise realization from provided'
            'power_spectal_density')
        frequency_domain_strain, frequency_array = \
            power_spectral_density.get_noise_realisation(
                self.sampling_frequency, self.duration)

        if np.array_equal(frequency_array, self.frequency_array):
            self._frequency_domain_strain = frequency_domain_strain
        else:
            raise ValueError("Data frequencies do not match frequency_array")

    def set_from_zero_noise(self, sampling_frequency, duration, start_time=0):
        """ Set the `frequency_domain_strain` to zero noise

        Parameters
        ==========
        sampling_frequency: float
            The sampling frequency (in Hz)
        duration: float
            The data duration (in s)
        start_time: float
            The GPS start-time of the data

        """

        self._times_and_frequencies = CoupledTimeAndFrequencySeries(duration=duration,
                                                                    sampling_frequency=sampling_frequency,
                                                                    start_time=start_time)
        logger.debug('Setting zero noise data')
        self._frequency_domain_strain = np.zeros_like(self.frequency_array,
                                                      dtype=complex)

    def set_from_frame_file(
            self, frame_file, sampling_frequency, duration, start_time=0,
            channel=None, buffer_time=1):
        """ Set the `frequency_domain_strain` from a frame fiile

        Parameters
        ==========
        frame_file: str
            File from which to load data.
        channel: str
            Channel to read from frame.
        sampling_frequency: float
            The sampling frequency (in Hz)
        duration: float
            The data duration (in s)
        start_time: float
            The GPS start-time of the data
        buffer_time: float
            Read in data with `start_time-buffer_time` and
            `start_time+duration+buffer_time`

        """

        self._times_and_frequencies = CoupledTimeAndFrequencySeries(
            duration=duration, sampling_frequency=sampling_frequency,
            start_time=start_time)

        logger.info('Reading data from frame file {}'.format(frame_file))
        strain = gwutils.read_frame_file(
            frame_file, start_time=start_time, end_time=start_time + duration,
            buffer_time=buffer_time, channel=channel,
            resample=sampling_frequency)

        self.set_from_gwpy_timeseries(strain)

    def set_from_channel_name(self, channel, duration, start_time, sampling_frequency):
        """ Set the `frequency_domain_strain` by fetching from given channel
        using gwpy.TimesSeries.get(), which dynamically accesses either frames
        on disk, or a remote NDS2 server to find and return data. This function
        also verifies that the specified channel is given in the correct format.

        Parameters
        ==========
        channel: str
            Channel to look for using gwpy in the format `IFO:Channel`
        duration: float
            The data duration (in s)
        start_time: float
            The GPS start-time of the data
        sampling_frequency: float
            The sampling frequency (in Hz)

        """
        from gwpy.timeseries import TimeSeries
        channel_comp = channel.split(':')
        if len(channel_comp) != 2:
            raise IndexError('Channel name must have format `IFO:Channel`')

        self._times_and_frequencies = CoupledTimeAndFrequencySeries(
            duration=duration, sampling_frequency=sampling_frequency,
            start_time=start_time)

        logger.info('Fetching data using channel {}'.format(channel))
        strain = TimeSeries.get(channel, start_time, start_time + duration)
        strain = strain.resample(sampling_frequency)

        self.set_from_gwpy_timeseries(strain)


class Notch(object):
    def __init__(self, minimum_frequency, maximum_frequency):
        """ A notch object storing the maximum and minimum frequency of the notch

        Parameters
        ==========
        minimum_frequency, maximum_frequency: float
            The minimum and maximum frequency of the notch

        """

        if 0 < minimum_frequency < maximum_frequency < np.inf:
            self.minimum_frequency = minimum_frequency
            self.maximum_frequency = maximum_frequency
        else:
            msg = ("Your notch minimum_frequency {} and maximum_frequency {} are invalid"
                   .format(minimum_frequency, maximum_frequency))
            raise ValueError(msg)

    def get_idxs(self, frequency_array):
        """ Get a boolean mask for the frequencies in frequency_array in the notch

        Parameters
        ==========
        frequency_array: np.ndarray
            An array of frequencies

        Returns
        =======
        idxs: np.ndarray
            An array of booleans which are True for frequencies in the notch

        """
        lower = (frequency_array > self.minimum_frequency)
        upper = (frequency_array < self.maximum_frequency)
        return lower & upper

    def check_frequency(self, freq):
        """ Check if freq is inside the notch

        Parameters
        ==========
        freq: float
            The frequency to check

        Returns
        =======
        True/False:
            If freq inside the notch, return True, else False
        """

        if self.minimum_frequency < freq < self.maximum_frequency:
            return True
        else:
            return False


class NotchList(list):
    def __init__(self, notch_list):
        """ A list of notches

        Parameters
        ==========
        notch_list: list
            A list of length-2 tuples of the (max, min) frequency for the
            notches.

        Raises
        ======
        ValueError
            If the list is malformed.
        """

        if notch_list is not None:
            for notch in notch_list:
                if isinstance(notch, tuple) and len(notch) == 2:
                    self.append(Notch(*notch))
                else:
                    msg = "notch_list {} is malformed".format(notch_list)
                    raise ValueError(msg)

    def check_frequency(self, freq):
        """ Check if freq is inside the notch list

        Parameters
        ==========
        freq: float
            The frequency to check

        Returns
        =======
        True/False:
            If freq inside any of the notches, return True, else False
        """

        for notch in self:
            if notch.check_frequency(freq):
                return True
        return False



def docstring(docstr, sep="\n"):
    """
    Decorator: Append to a function's docstring.

    This is required for e.g., :code:`classmethods` as the :code:`__doc__`
    can't be changed after.

    Parameters
    ==========
    docstr: str
        The docstring
    sep: str
        Separation character for appending the existing docstring.
    """
    def _decorator(func):
        if func.__doc__ is None:
            func.__doc__ = docstr
        else:
            func.__doc__ = sep.join([func.__doc__, docstr])
        return func
    return _decorator



import unittest
import bilby

import numpy as np


class TestMultivariateGaussianDistFromRepr(unittest.TestCase):
    def test_mvg_from_repr(self):
        mvg = bilby.core.prior.MultivariateGaussianDist(
            names=["testa", "testb"],
            mus=[1, 1],
            covs=np.array([[2.0, 0.5], [0.5, 2.0]]),
            weights=1.0,
        )

        # string representation
        mvgstr = """\
MultivariateGaussianDist(
    names=['testa', 'testb'],
    nmodes=1,
    mus=[[1, 1]],
    corrcoefs=[[[1.0, 0.25], [0.25, 1.0]]],
    covs=[[[2.0, 0.5], [0.5, 2.0]]],
    weights=[1.0],
    bounds={'testa': (-inf, inf), 'testb': (-inf, inf)}
)"""

        fromstr = bilby.core.prior.MultivariateGaussianDist.from_repr(mvgstr)

        for key, item in mvg.__dict__.items():
            if isinstance(item, dict):
                self.assertTrue(item == fromstr.__getattribute__(key))
            elif key == "mvn":
                for d1, d2 in zip(fromstr.__getattribute__(key), item):
                    self.assertTrue(type(d1) == type(d2))
            elif isinstance(item, (list, tuple, np.ndarray)):
                self.assertTrue(
                    np.all(np.array(item) == np.array(fromstr.__getattribute__(key)))
                )


if __name__ == "__main__":
    unittest.main()



import torch
from nflows.distributions.normal import StandardNormal
from nflows.flows.base import Flow
from nflows.nn import nets as nets
from nflows.transforms import (
    CompositeTransform,
    MaskedAffineAutoregressiveTransform,
    RandomPermutation,
)
from nflows.transforms.coupling import (
    AdditiveCouplingTransform,
    AffineCouplingTransform,
)
from nflows.transforms.normalization import BatchNorm
from torch.nn import functional as F

# Turn off parallelism
torch.set_num_threads(1)
torch.set_num_interop_threads(1)


class NVPFlow(Flow):
    """A simplified version of Real NVP for 1-dim inputs.

    This implementation uses 1-dim checkerboard masking but doesn't use
    multi-scaling.
    Reference:
    > L. Dinh et al., Density estimation using Real NVP, ICLR 2017.

    This class has been modified from the example found at:
    https://github.com/bayesiains/nflows/blob/master/nflows/flows/realnvp.py
    """

    def __init__(
        self,
        features,
        hidden_features,
        num_layers,
        num_blocks_per_layer,
        use_volume_preserving=False,
        activation=F.relu,
        dropout_probability=0.0,
        batch_norm_within_layers=False,
        batch_norm_between_layers=False,
        random_permutation=True,
    ):

        if use_volume_preserving:
            coupling_constructor = AdditiveCouplingTransform
        else:
            coupling_constructor = AffineCouplingTransform

        mask = torch.ones(features)
        mask[::2] = -1

        def create_resnet(in_features, out_features):
            return nets.ResidualNet(
                in_features,
                out_features,
                hidden_features=hidden_features,
                num_blocks=num_blocks_per_layer,
                activation=activation,
                dropout_probability=dropout_probability,
                use_batch_norm=batch_norm_within_layers,
            )

        layers = []
        for _ in range(num_layers):
            transform = coupling_constructor(
                mask=mask, transform_net_create_fn=create_resnet
            )
            layers.append(transform)
            mask *= -1
            if batch_norm_between_layers:
                layers.append(BatchNorm(features=features))

        if random_permutation:
            layers.append(RandomPermutation(features=features))

        super().__init__(
            transform=CompositeTransform(layers),
            distribution=StandardNormal([features]),
        )


class BasicFlow(Flow):
    def __init__(self, features):
        transform = CompositeTransform(
            [
                MaskedAffineAutoregressiveTransform(
                    features=features, hidden_features=2 * features
                ),
                RandomPermutation(features=features),
            ]
        )
        distribution = StandardNormal(shape=[features])
        super().__init__(
            transform=transform,
            distribution=distribution,
        )



import numpy as np


def ra_dec_to_theta_phi(ra, dec, gmst):
    """ Convert from RA and DEC to polar coordinates on celestial sphere

    Parameters
    ==========
    ra: float
        right ascension in radians
    dec: float
        declination in radians
    gmst: float
        Greenwich mean sidereal time of arrival of the signal in radians

    Returns
    =======
    float: zenith angle in radians
    float: azimuthal angle in radians

    """
    phi = ra - gmst
    theta = np.pi / 2 - dec
    return theta, phi


def theta_phi_to_ra_dec(theta, phi, gmst):
    ra = phi + gmst
    dec = np.pi / 2 - theta
    return ra, dec



import unittest
import numpy as np
import pandas as pd
from parameterized import parameterized

import bilby.hyper as hyp


def _toy_function(data, dataset, self, cls, a, b, c):
    return a


class _ToyClassNoVariableNames:
    def __call__(self, a, b, c):
        return a


class _ToyClassVariableNames:

    variable_names = ["a", "b", "c"]

    def __call__(self, **kwargs):
        return kwargs.get("a", 1)


class TestHyperLikelihood(unittest.TestCase):
    def setUp(self):
        self.keys = ["a", "b", "c"]
        self.lengths = [300, 400, 500]
        self.posteriors = list()
        for ii, length in enumerate(self.lengths):
            self.posteriors.append(
                pd.DataFrame({key: np.random.normal(0, 1, length) for key in self.keys})
            )
        self.log_evidences = [2, 2, 2]
        self.model = hyp.model.Model(list())
        self.sampling_model = hyp.model.Model(list())

    def tearDown(self):
        del self.keys
        del self.lengths
        del self.posteriors
        del self.log_evidences

    def test_evidence_factor_with_evidences(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors,
            self.model,
            self.sampling_model,
            log_evidences=self.log_evidences,
        )
        self.assertEqual(like.evidence_factor, 6)

    def test_evidence_factor_without_evidences(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors, self.model, self.sampling_model
        )
        self.assertTrue(np.isnan(like.evidence_factor))

    @parameterized.expand([
        ("func", _toy_function),
        ("class_no_names", _ToyClassNoVariableNames()),
        ("class_with_names", _ToyClassVariableNames()),
    ])
    def test_get_function_parameters(self, _, model):
        expected = dict(a=1, b=2, c=3)
        model = hyp.model.Model([model])
        model.parameters.update(expected)
        result = model._get_function_parameters(model.models[0])
        self.assertDictEqual(expected, result)

    def test_len_samples_with_max_samples(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors,
            self.model,
            self.sampling_model,
            log_evidences=self.log_evidences,
            max_samples=10,
        )
        self.assertEqual(like.samples_per_posterior, 10)

    def test_len_samples_without_max_samples(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors,
            self.model,
            self.sampling_model,
            log_evidences=self.log_evidences,
        )
        self.assertEqual(like.samples_per_posterior, min(self.lengths))

    def test_resample_with_max_samples(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors,
            self.model,
            self.sampling_model,
            log_evidences=self.log_evidences,
        )
        resampled = like.resample_posteriors()
        self.assertEqual(resampled["a"].shape, (len(self.lengths), min(self.lengths)))

    def test_resample_without_max_samples(self):
        like = hyp.likelihood.HyperparameterLikelihood(
            self.posteriors,
            self.model,
            self.sampling_model,
            log_evidences=self.log_evidences,
        )
        resampled = like.resample_posteriors(10)
        self.assertEqual(resampled["a"].shape, (len(self.lengths), 10))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from setuptools import setup
import subprocess
import sys
import os

python_version = sys.version_info
if python_version < (3, 8):
    sys.exit("Python < 3.8 is not supported, aborting setup")


def write_version_file(version):
    """Writes a file with version information to be used at run time

    Parameters
    ----------
    version: str
        A string containing the current version information

    Returns
    -------
    version_file: str
        A path to the version file

    """
    try:
        git_log = subprocess.check_output(
            ["git", "log", "-1", "--pretty=%h %ai"]
        ).decode("utf-8")
        git_diff = (
            subprocess.check_output(["git", "diff", "."])
            + subprocess.check_output(["git", "diff", "--cached", "."])
        ).decode("utf-8")
        if git_diff == "":
            git_status = "(CLEAN) " + git_log
        else:
            git_status = "(UNCLEAN) " + git_log
    except Exception as e:
        print("Unable to obtain git version information, exception: {}".format(e))
        git_status = "release"

    version_file = ".version"
    if os.path.isfile(version_file) is False:
        with open("bilby/" + version_file, "w+") as f:
            f.write("{}: {}".format(version, git_status))

    return version_file


def get_long_description():
    """Finds the README and reads in the description"""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, "README.rst")) as f:
        long_description = f.read()
    return long_description


def get_requirements(kind=None):
    if kind is None:
        fname = "requirements.txt"
    else:
        fname = f"{kind}_requirements.txt"
    with open(fname, "r") as ff:
        requirements = ff.readlines()
    return requirements


# get version info from __init__.py
def readfile(filename):
    with open(filename) as fp:
        filecontents = fp.read()
    return filecontents


VERSION = '1.2.0'
version_file = write_version_file(VERSION)
long_description = get_long_description()

setup(
    name="bilby",
    description="A user-friendly Bayesian inference library",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://git.ligo.org/lscsoft/bilby",
    author="Greg Ashton, Moritz Huebner, Paul Lasky, Colm Talbot",
    author_email="paul.lasky@monash.edu",
    license="MIT",
    version=VERSION,
    packages=[
        "bilby",
        "bilby.bilby_mcmc",
        "bilby.core",
        "bilby.core.prior",
        "bilby.core.sampler",
        "bilby.core.utils",
        "bilby.gw",
        "bilby.gw.detector",
        "bilby.gw.eos",
        "bilby.gw.likelihood",
        "bilby.gw.sampler",
        "bilby.hyper",
        "cli_bilby",
    ],
    package_dir={"bilby": "bilby", "cli_bilby": "cli_bilby"},
    package_data={
        "bilby.gw": ["prior_files/*"],
        "bilby.gw.detector": ["noise_curves/*.txt", "detectors/*"],
        "bilby.gw.eos": ["eos_tables/*.dat"],
        "bilby": [version_file],
    },
    python_requires=">=3.8",
    install_requires=get_requirements(),
    extras_require={
        "gw": get_requirements("gw"),
        "mcmc": get_requirements("mcmc"),
        "all": (
            get_requirements("sampler")
            + get_requirements("gw")
            + get_requirements("mcmc")
            + get_requirements("optional")
        ),
    },
    entry_points={
        "console_scripts": [
            "bilby_plot=cli_bilby.plot_multiple_posteriors:main",
            "bilby_result=cli_bilby.bilby_result:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import sys

import bilby  # noqa

unique_packages = set(sys.modules)

unwanted = {
    "lal", "lalsimulation", "matplotlib",
    "h5py", "dill", "tqdm", "tables", "deepdish", "corner",
}

for filename in ["sampler_requirements.txt", "optional_requirements.txt"]:
    with open(filename, "r") as ff:
        packages = ff.readlines()
        for package in packages:
            package = package.split(">")[0].split("<")[0].split("=")[0].strip()
            unwanted.add(package)

if not unique_packages.isdisjoint(unwanted):
    raise ImportError(
        f"{' '.join(unique_packages.intersection(unwanted))} imported with Bilby"
    )



from .calculus import *
from .cmd import *
from .colors import *
from .constants import *
from .conversion import *
from .counter import *
from .docs import *
from .introspection import *
from .io import *
from .log import *
from .plotting import *
from .samples import *
from .series import *

#  Instantiate the default argument parser at runtime
command_line_args, command_line_parser = set_up_command_line_arguments()
#  Instantiate the default logging
setup_logger(print_version=False, log_level=command_line_args.log_level)
